/*******
 
 InkWeb v0.8
 By Rayhan Hasan
 
 --with inspiration from Mr. Doob's "Harmony"
 available at http://mrdoob.com/projects/harmony/#web
 
 --Mouse input smoothing algorithm based on 
 work by Paul Nettle, available at
 http://www.flipcode.com/archives/Smooth_Mouse_Filtering.shtml
 
 --Special thanks to sojamo, who developed controlP5
 and made things easier to control.
 
 --This code could use a thorough refactoring, but that's
 a project for another day.
 
 Licensed to you under the GNU GPL v3
 
 You know.
 

 
 TODO:
 -handle resizes more gracefully
 
 -explore bezier curves as a replacement for straight lines
 -set up hiding/showing of controls (and figure out a nice 
 way of telling users that hitting spacebar brings it up.
 -try to produce work, rework code to facilitate
 *******/
 
import controlP5.*;
import java.awt.geom.Point2D;
import java.util.LinkedList;
import java.util.Queue;

//Defaults
double MAX_STROKE = 3;
double MIN_STROKE = 1;

int WEB_SPARSENESS = 5;
double MAX_WEB_LENGTH = 90;
double WEB_OPACITY = 15;

//These are the variables used throughout the program
//They are set to the default values in setupControllers()
int lineColor;
double maxStroke;
double minStroke;


int webSparseness;
double maxWebLength;
double webOpacity;

boolean mouseIsDown;

int cX, cY, pX, pY;
int offset;
int numSaves=0;
boolean changedSinceLastSave;

boolean enableStroke;
boolean enableWebs;




LinkedList<Point2D.Double> listOfPoints;
LinkedList<Point2D.Double> smoothingBuffer;

ControlP5 controlP5;
ControlWindow controlWindow;

/*
Toggle linesToggle;
Toggle websToggle;
*/

///todo: add buttons for enable webs and enable stroke (also sync them up with vars and keypresses)


void setup() {
  size(700,700);
  frame.setResizable(true);
  smooth();
  colorMode(RGB, 100);
  background(100);
  cursor(CROSS);

  strokeJoin(ROUND);
  strokeCap(ROUND);
  rectMode(CENTER);
  enableStroke=true;
  enableWebs=true;
  clearBoard();
  
  setupControllers();



  //smoothingBuffer is always working, even when not drawing.
  //this is why it's not reset in forgetPoints()
  smoothingBuffer =  new LinkedList<Point2D.Double>();
}

void clearBoard() {
  background(100);
  changedSinceLastSave=false;
  forgetPoints();
}

void forgetPoints() {
  listOfPoints = new LinkedList<Point2D.Double>();
  offset=0;
}

void setupControllers() {
  maxStroke=MAX_STROKE;
  minStroke=MIN_STROKE;
  
  webSparseness = WEB_SPARSENESS;  
  maxWebLength = MAX_WEB_LENGTH;
  webOpacity = WEB_OPACITY;

  controlP5 = new ControlP5(this);
  
  controlWindow = controlP5.addControlWindow("controlP5window",100,100,330,180);
  controlWindow.hideCoordinates();
  controlWindow.setBackground(color(100));
  controlWindow.setTitle("InkWeb Settings");
  //controlWindow.frameRate(60);
 
  int startingHeight=15;
  int gap = 22;

/*  
  linesToggle = controlP5.addToggle("Ink",true,20,startingHeight,20,20);
  linesToggle.setColorLabel(0x0);
  linesToggle.moveTo(controlWindow);
  
  
  websToggle = controlP5.addToggle("Webs",true,70,startingHeight,20,20);
  websToggle.setColorLabel(0x0);
  websToggle.moveTo(controlWindow);
*/

  Range lineWidthRange = controlP5.addRange("Ink Thickness",0,9,(float)MIN_STROKE,(float)MAX_STROKE, 20,(int)(startingHeight),200,14);
  lineWidthRange.setColorLabel(0x0);
  lineWidthRange.moveTo(controlWindow);


  Slider webOpSlider = controlP5.addSlider("Web Opacity", 0, 100, (float)WEB_OPACITY, 20, (int)(startingHeight+gap*1), 200, 14);
  webOpSlider.setColorLabel(0x0);
  webOpSlider.moveTo(controlWindow);
  
  Slider webSpa = controlP5.addSlider("Web Density", 1, 25, (float)WEB_SPARSENESS, 20, (int)(startingHeight+gap*2), 200, 14);
  webSpa.setColorLabel(0x0);
  webSpa.moveTo(controlWindow);
  
  Slider webLen = controlP5.addSlider("Web Length", 10, 1000, (float)MAX_WEB_LENGTH, 20, (int)(startingHeight+gap*3), 200, 14);
  webLen.setColorLabel(0x0);
  webLen.moveTo(controlWindow);
    
  Textarea hotkeyInfo = controlP5.addTextarea("Hotkey Info", "Spacebar: Save & clear   S: Save                        "+
  "Q: Toggle Ink   W: Toggle Webs   F: Forget web points                                                            "+
  "InkWeb v0.8 by rayhan hasan. http://dayvancowboy.com/code",
      20, (int)(startingHeight+gap*5), 280, 40);
  hotkeyInfo.moveTo(controlWindow);
  hotkeyInfo.setColor(0x0);
    
}
void draw() {
  smoothMouse(mouseX,mouseY); // sets cX, cY, pX, pY 
  
  if(mouseIsDown)
    drawLines();

  offset += (int) random(4,6);
}


void drawLines() {

  Point2D.Double currentPoint=new Point2D.Double(cX, cY);
  Point2D.Double previousPoint=new Point2D.Double(pX, pY);


  if(listOfPoints.indexOf(currentPoint) == -1) {//the point is not already plotted

    if(enableWebs==true) {//start draw webs
      for(int i=0;i<listOfPoints.size();i++) {

        //println(listOfPoints.size());
        if((i+offset)%(26-webSparseness)==0 && currentPoint.distance(listOfPoints.get(i)) < maxWebLength) {


          stroke(0,0,0,(float)webOpacity);
          strokeWeight(1);

          line((int)currentPoint.getX(),(int) currentPoint.getY(),
          (int) listOfPoints.get(i).getX(),(int) listOfPoints.get(i).getY());
        }
      }
    } //end draw webs

    listOfPoints.add(currentPoint);
  }


  if(enableStroke) {
    stroke(0);
    strokeWeight(calcStrokeWeight(currentPoint, previousPoint));

    line((int)currentPoint.getX(), (int)currentPoint.getY(), 
    (int)previousPoint.getX(),(int)previousPoint.getY());
  }
  
  changedSinceLastSave = true;
}


float calcStrokeWeight(Point2D.Double currentPoint, Point2D.Double previousPoint) {

  double newStroke = currentPoint.distance(previousPoint)/5;

  if(newStroke > maxStroke)
    newStroke=maxStroke;

  newStroke=maxStroke-newStroke;

  if(newStroke < minStroke)
    newStroke=minStroke;

  return (float)newStroke;
}


void keyReleased() {
  
  if( key == 'q' || key=='Q') {
    enableStroke=!enableStroke;
    //when toggle controls are written, update
    //the controller's state here.
  }
  
  
  else if( key == 'w' || key=='W') {
    enableWebs=!enableWebs;
    //when toggle controls are written, update
    //the controller's state here.
  }
  
  
  
}



void keyPressed() {
  if( key == 'f' || key=='F') {
    forgetPoints();
  }



  if( key == ' ') {

    if(changedSinceLastSave)
      save(dataPath(findNextFilename()));

    clearBoard();
    changedSinceLastSave = false; //already in setup, but here for clarity
  }

  if( (key == 's' || key == 'S') && changedSinceLastSave) {

    save(dataPath(findNextFilename()));
    changedSinceLastSave = false;
  }
}


void mousePressed() {

  if(mouseButton == LEFT)
    mouseIsDown=true;


  else if(mouseButton == RIGHT) { 
    //   forgetPoints();
  }
}

void mouseDragged() {
}

void mouseReleased() {
  mouseIsDown=false;
}


void smoothMouse(int x, int y) {

  if(smoothingBuffer.size() > 10)
    smoothingBuffer.removeLast();

  smoothingBuffer.addFirst(new Point2D.Double((double)x, (double)y));

  Point2D.Double newPoint = getFilteredPoint();

  pX=cX;
  pY=cY;
  cX=(int)newPoint.getX();
  cY=(int)newPoint.getY();
}

Point2D.Double getFilteredPoint() {
  //the weights add up to 2 (more or less)
  double weight;
  double sum;
  double xVal, yVal;

  weight=1.0;
  sum=0;
  for(int i=0;i<smoothingBuffer.size();i++) {
    sum= sum + ((smoothingBuffer.get(i).getX())*(weight));
    weight = weight/2;
  }
  xVal=sum/2;

  weight=1.0;
  sum=0;
  for(int i=0;i<smoothingBuffer.size();i++) {
    sum= sum + (smoothingBuffer.get(i).getY())*(weight);
    weight=weight/2;
  }
  yVal=sum/2;


  return new Point2D.Double(xVal,yVal);
}


void controlEvent(ControlEvent theEvent) {
  // println("got a control event from controller with id "+theEvent.controller().id());


/*
  if(theEvent.controller().name() == "Ink") {
    enableStroke=!enableStroke;
    
  }
  else if(theEvent.controller().name() == "Webs") {
    enableWebs=!enableWebs;
  
  }
*/


  if(theEvent.controller().name() == "Ink Thickness") {
    minStroke = theEvent.arrayValue()[0]; 
    maxStroke = theEvent.arrayValue()[1];
  } 

    
  else if(theEvent.controller().name() == "Web Density") {
    webSparseness=(int)theEvent.value();
  }
  else if(theEvent.controller().name() == "Web Length")
    maxWebLength=(double)theEvent.value();
  else if(theEvent.controller().name() == "Web Opacity")
    webOpacity=(double)theEvent.value();
}

String findNextFilename() {

  int saveNumber=0;
  java.text.DecimalFormat twoDecPlaces = new java.text.DecimalFormat("00");
  java.text.DecimalFormat fourDecPlaces = new java.text.DecimalFormat("0000");


  String filename= "InkWeb-"+year()+"-"+twoDecPlaces.format(month())+"-"+
    twoDecPlaces.format(day())+"-"+fourDecPlaces.format(saveNumber)+".png";

  File f = new File(dataPath(filename));

  while(f.exists()) {
    saveNumber++;
    filename= "InkWeb-"+year()+"-"+twoDecPlaces.format(month())+"-"+
      twoDecPlaces.format(day())+"-"+fourDecPlaces.format(saveNumber)+".png";
    f = new File(dataPath(filename));
  }

  return filename;
}




void stop()
{ 
  if(changedSinceLastSave) {
    save(dataPath(findNextFilename()));
  }
} 

